/*
 * Capsule
 * Copyright (c) 2014-2015, Parallel Universe Software Co. All rights reserved.
 * 
 * This program and the accompanying materials are licensed under the terms 
 * of the Eclipse Public License v1.0, available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
/**
 * Contains classes to interact with capsules:
 * 
 * <ul>
 * <li>{@link co.paralleluniverse.capsule.CapsuleUtils} lets an application running in a capsule obtain information from the capsule.</li>
 * <li>{@link co.paralleluniverse.capsule.Jar} allows simple, programmatic creation of JAR file/streams, and can be used by a server that generates customized capsules.</li>
 * <li>{@link co.paralleluniverse.capsule.CapsuleLauncher} launched capsules programmatically, and is used by capsule containers.</li>
 * </ul>
 */
package co.paralleluniverse.capsule;
